#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
""" This is the CARTA launcher. """
import logging
import math
import os
import signal
import subprocess
import sys
import time
from http.server import HTTPServer, SimpleHTTPRequestHandler
from pathlib import Path
from threading import Thread
from types import FrameType
from typing import Optional

import pendulum as pendulum
from carta_envoy.connect import (
    ArchiveConnect,
    NotificationConnect,
    RedisConnect,
    WorkflowConnect,
)
from carta_envoy.templates import CartaWrapper
from pendulum import DateTime

# pylint: disable=E0401, R1705, W0105


# pylint: disable=E0401, R0913, R1721, W0603, W1203

"""
Setup and Launch a CARTA session
"""

logger = logging.getLogger("carta_envoy")
CARTA_PROCESS: Optional[subprocess.Popen] = None


class CartaWrapperLauncher:
    """Serves an HTML wrapper for CARTA"""

    @staticmethod
    def run(port: int, server_class=HTTPServer, handler_class=SimpleHTTPRequestHandler):
        """
        Run simple web server for hosting CARTA wrapper HTML

        :param port: Port to run the server on
        :param server_class: The class that handles serverly duties
        :param handler_class: The class that handles HTTP requests to the server
        """
        server_address = ("", port)
        with server_class(server_address, handler_class) as httpd:
            httpd.serve_forever()

    @staticmethod
    def deploy_wrapper_html(product_dir: Path, carta_url: str, session_timeout_date: str):
        """
        Deploy wrapper HTML to product directory as `wrapper.html`

        :param product_dir: Path to deploy HTML to
        :param carta_url: URL for CARTA instance
        :param session_timeout_date: Datetime at which this CARTA session will expire
        """
        wrapper_html = CartaWrapper().render_template(carta_url, session_timeout_date)

        # Deploy filled-out wrapper HTML to product directory
        with open(product_dir / "index.html", "w") as complete_wrapper_file:
            complete_wrapper_file.write(wrapper_html)


class CartaLauncher:
    """Runs CARTA"""

    def __init__(self, settings: dict):
        self.logger = logging.getLogger("carta_envoy")
        self.settings = settings
        self.signal = self.signal_handler
        self.redis_connect = RedisConnect(settings)
        self.archive_connect = ArchiveConnect(settings)
        self.workflow_connect = WorkflowConnect(settings)
        self.notification = NotificationConnect(settings)

    def signal_handler(self, signum: int, frame: FrameType):
        """
        Handle a signal received by tearing down the program and exiting
        NOTE: Parameters signum and frame are convention for signal handlers,
        even if they don't end up used

        :param signum: Signal number
        :param frame: Stack frame
        """
        if signum == signal.SIGINT:
            self.logger.warning("WARNING: SIGINT received. Exiting.")
        elif signum == signal.SIGALRM:
            self.logger.warning("WARNING: Timeout elapsed. Exiting.")
        self.teardown()
        sys.exit()

    def prepare_for_carta(self):
        """
        Sets Redis values necessary for CARTA.

        :return:
        """
        ports_and_url = self.redis_connect.prepare_redis()
        signal.signal(signal.SIGINT, self.signal)
        return ports_and_url

    def launch(self):
        """
        Launches CARTA and its wrapper.
        """
        self.logger.info("RUNNING CARTA WITH WRAPPER!")
        prepare = self.prepare_for_carta()
        session_timeout_date = str(self._calculate_timeout_date(int(self.settings["timeout"])))

        self.run_carta(
            self.settings["carta_path"],
            int(self.settings["timeout"]),
            Path(self.settings["data_location"]),
            prepare["session_port"],
            prepare["wrapper_port"],
            prepare["carta_url"],
            prepare["wrapper_url"],
            session_timeout_date,
        )

    def run_carta(
        self,
        path_to_carta: str,
        timeout_minutes: int,
        file_browser_path: Path,
        session_port: int,
        wrapper_port: int,
        carta_url: str,
        wrapper_url: str,
        session_timeout_date: str,
    ):
        """
        Runs CARTA executable. Will close after a certain amount of time.

        :param path_to_carta: Path to CARTA executable
        :param timeout_minutes: Number of minutes to execute CARTA for before stopping it
        :param file_browser_path: Path to use for CARTA file browser
        :param session_port: Port to run the CARTA session with
        :param wrapper_port: Port to run the CARTA wrapper with
        :param carta_url: URL used to access CARTA from a web browser
        :param wrapper_url: URL used to access CARTA wrapper from a web browser
        :param session_timeout_date: Date at which CARTa session will expire
        """
        global CARTA_PROCESS

        self.logger.info(f"file_browser_path: {file_browser_path}")
        if not file_browser_path.exists():
            self.logger.error(f"ERROR: Invalid file browser path. {file_browser_path}: Directory does not exist.")
            return
        elif not file_browser_path.is_dir():
            self.logger.error(f"ERROR: Invalid file browser path. {file_browser_path}: Path is not a directory.")
            return

        if self.settings["useCarta"]:
            self.logger.info(f"Running CARTA with a timeout of {timeout_minutes} minute(s)...")
            carta_command = [
                path_to_carta,
                "--no_browser",  # 2.0 setting: don't automatically try to open browser
                "--debug_no_auth",  # No tokenized authorization
                "--no_runtime_config",  # Don't send runtime config object to CARTA front-end (fixes a bug)
                f"--port={session_port!s}",
                f"--top_level_folder={file_browser_path!s}",
                f"{file_browser_path!s}",
            ]
            try:
                # finish setup for CARTA and ensure completion before launch
                CartaWrapperLauncher.deploy_wrapper_html(file_browser_path, carta_url, session_timeout_date)
                time.sleep(3)

                #  start CARTA
                CARTA_PROCESS = subprocess.Popen(
                    carta_command,
                    preexec_fn=os.setpgrp(),
                    stdin=subprocess.PIPE,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                )
            except (OSError, FileNotFoundError, subprocess.SubprocessError) as err:
                self.teardown()
                sys.exit(f"ERROR: Failed to launch CARTA: {err}")
            else:
                # CartaWrapperLauncher.deploy_wrapper_html(file_browser_path, carta_url, session_timeout_date)

                # CARTA is running and accessible, so send CARTA URL to AAT system or notify user
                self.notify_ready(wrapper_url)

                # Activate timeout handler
                signal.signal(signal.SIGALRM, self.signal)
                signal.alarm(60 * timeout_minutes)
                self.logger.info(f"To access CARTA, enter this URL into a web browser: {wrapper_url}")
                self.logger.info("Press Ctrl+C to quit.")

                wrapper_server_thread = Thread(target=CartaWrapperLauncher.run, args=[wrapper_port], daemon=True)
                wrapper_server_thread.run()

                # Loop until process ends or timeout elapses, whichever happens first
                while CARTA_PROCESS.returncode is None:
                    CARTA_PROCESS.poll()
                    assert wrapper_server_thread.is_alive()
        else:
            self.logger.info("Running locally...")

    def notify_ready(self, wrapper_url: str):
        """
        Sends URL notification to user and request handler

        :param wrapper_url: URL to CARTA session
        """

        self.logger.info("SENDING URL NOTIFICATION TO: ")
        if self.settings["send_ready"] == "true":
            self.logger.info(f"Workspaces, with {wrapper_url}")
            # WS-1093: DAs requested deactivation of email sending for WS CARTA
            # self.notification.send_session_ready(wrapper_url)
            self.workflow_connect.send_carta_url(wrapper_url)
        elif self.settings["send_ready"] == "false":
            self.logger.info(f"AAT Request Handler, with {wrapper_url}")
            self.archive_connect.send_carta_url_to_rh(wrapper_url)

    def teardown(self):
        """
        Deletes instance's Redis keys from the Redis server
        """
        self.notification.send_session_expired()

        self.logger.info("Cleaning up...")
        self.redis_connect.remove_redis_values()

        if CARTA_PROCESS and CARTA_PROCESS.returncode is None:
            # If CARTA is running, kill it
            try:
                # End the process (simulate Ctrl+Z)
                CARTA_PROCESS.send_signal(signal.SIGSTOP)
                # Kill the process
                os.killpg(os.getpgid(CARTA_PROCESS.pid), signal.SIGTERM)
                CARTA_PROCESS.wait()
            except OSError:
                # Process isn't running; just pass
                pass
            else:
                self.logger.info("Killed running CARTA process.")
        else:
            self.logger.warning("WARNING: CARTA not running.")

    @staticmethod
    def _calculate_timeout_date(timeout_minutes: int) -> DateTime:
        """
        Take the timeout of the CARTA session amd use it to calculate the datetime that the session will expire at

        :param timeout_minutes: Timeout of the CARTA session
        :return: Datetime at which CARTA session will expire
        """
        logger.info(f"Creating timestamp for end of timeout ({timeout_minutes} minutes)")
        if timeout_minutes > 59:
            timeout_hours, timeout_remainder = math.modf(timeout_minutes / 60)
            timeout_hours = int(timeout_hours)
            timeout_minutes = int(timeout_remainder * 60)
            logger.info(f"Timeout longer than 1 hour. Splitting into {timeout_hours}:{timeout_minutes}")
            timeout_date = pendulum.now().add(hours=timeout_hours, minutes=timeout_minutes)
        else:
            timeout_date = pendulum.now().add(minutes=timeout_minutes)

        logger.info(f"Final timeout date of {timeout_date}")
        return timeout_date
